package net.pingex.dcf.commands;

import net.pingex.dcf.commands.options.ICommandOption;
import net.pingex.dcf.commands.options.ScopeOption;
import net.pingex.dcf.commands.permissions.DefaultPermissionOption;
import net.pingex.dcf.core.Configuration;
import net.pingex.dcf.util.ArgumentParser;
import net.pingex.dcf.util.DiscordInteractionsUtil;
import org.apache.commons.lang3.StringUtils;
import sx.blah.discord.handle.obj.IRole;
import java.util.*;

/**
 * Internal commands of DCF.
 */
public class InternalCommands implements IWithCommands
{
	@Override
	public Set<Command> getCommands()
	{
		return new HashSet<>(Arrays.asList(ListCommand.INSTANCE, UsageCommand.INSTANCE, dumpRoles));
	}

	/**
	 * This command list all available commands on DCF.
	 */
	private static class ListCommand extends Command
	{
		private static final String NAME = "internal:list";
		private static final List<String> ALIASES = Arrays.asList("list", "help");
		private static final String DESCRIPTION = "List all available commands.";
		private static final boolean IS_ENABLED = true;
		private static final String USAGE = "Page";
		private static final Set<ICommandOption> OPTIONS = Collections.singleton(new ScopeOption(ScopeOption.CommandScope.ANYWHERE));

		/**
		 * How many commands should be displayed on each page
		 */
		public static final int COMMANDS_PER_PAGE = 5;

		static final ListCommand INSTANCE = new ListCommand();

		private ListCommand()
		{
			super(NAME, ALIASES, DESCRIPTION, IS_ENABLED, USAGE, OPTIONS);
		}

		@Override
		public void execute(Context context)
		{
			// Parameters
			Set<Command> bank = CommandRegistry.getRegistry();

			int amountPages = (int) Math.ceil(bank.size()/(double)COMMANDS_PER_PAGE);
			int requestedPage;
			int longestCommand = bank.stream().max((o1, o2) -> Integer.compare(o1.getName().length(), o2.getName().length())).get().getName().length();
			int longestDesc = bank.stream().max((o1, o2) -> Integer.compare(o1.getDescription().length(), o2.getDescription().length())).get().getDescription().length();

			// Parsing
			try
			{
				List<Object> output = ArgumentParser.parseAll(Collections.singletonList(Integer.class), context.getArguments());
				requestedPage = output.get(0) != null ? (int) output.get(0) : 1;
			}
			catch(ArgumentParser.ParserException e)
			{
				DiscordInteractionsUtil.sendMessage(context.getChannel(), e.getMessage());
				return;
			}

			// Checks
			if(requestedPage <= 0 || requestedPage > amountPages)
			{
				DiscordInteractionsUtil.sendMessage(context.getChannel(), "Requested page is invalid. Number of available pages: " + amountPages);
				return;
			}

			StringBuilder output = new StringBuilder
					("**List of commands available for " + Configuration.BOT_NAME + "** (page " + requestedPage + "/" + amountPages + ")\n");
			output.append("Use `").append(UsageCommand.NAME).append("`")
					.append(" to get details and usage about any command.\n");
			output.append("```");

			int pos = 1;
			for(Command i : bank)
			{
				if(pos > (requestedPage-1)*COMMANDS_PER_PAGE && pos <= requestedPage*COMMANDS_PER_PAGE)
				{
					output.append("+ ").append(StringUtils.rightPad(i.getName(), longestCommand));								// Name
					output.append("   ").append(StringUtils.rightPad(i.getDescription(), longestDesc));							// Desc
					if(i.getAliases().size() > 0)
						output.append("   ").append("(aliases: ").append(StringUtils.join(i.getAliases(), ", ")).append(")");	// Aliases

					output.append("\n");
				}
				pos++;
			}

			output.append("```");
			DiscordInteractionsUtil.sendMessage(context.getChannel(), output.toString());
		}
	}

	/**
	 * Gives the usage of a command.
	 */
	private static class UsageCommand extends Command
	{
		private static final String NAME = "internal:usage";
		private static final List<String> ALIASES = Collections.singletonList("usage");
		private static final String DESCRIPTION = "Gives the usage of a command.";
		private static final boolean IS_ENABLED = true;
		private static final String USAGE = "Command";
		private static final Set<ICommandOption> OPTIONS = Collections.singleton(new ScopeOption(ScopeOption.CommandScope.ANYWHERE));

		static final UsageCommand INSTANCE = new UsageCommand();

		private UsageCommand()
		{
			super(NAME, ALIASES, DESCRIPTION, IS_ENABLED, USAGE, OPTIONS);
		}

		@Override
		public void execute(Context context) throws Throwable
		{
			// Checks
			if(context.getArguments().size() != 1) // Arg check
			{
				DiscordInteractionsUtil.sendMessage(context.getChannel(), "Invalid argument.");
				return;
			}

			Optional<Command> uncheckedTarget = CommandRegistry.getCommandOrAliasByName(context.getArguments().get(0));
			if(!uncheckedTarget.isPresent()) // Command existence
			{
				DiscordInteractionsUtil.sendMessage(context.getChannel(), "Target command not found.");
				return;
			}

			// Print everything
			Command target = uncheckedTarget.get();
			StringBuilder output = new StringBuilder();
			output.append("**Details and usage for command `")
					.append(Configuration.COMMAND_PREFIX).append(target.getName())
					.append("`**\n")
					.append("```");

			output.append("Description: ").append(target.getDescription()).append("\n");
			if(target.getAliases().size() > 0)
				output.append("Aliases:     ").append(StringUtils.join(target.getAliases(), ", ")).append("\n");
			output.append("Usage:       ").append(Configuration.COMMAND_PREFIX).append(target.getName()).append(" ").append(target.getUsage()).append("\n");
			output.append("Enabled:     ").append(target.isEnabled() ? "Yes" : "No").append("\n");

			output.append("```");
			DiscordInteractionsUtil.sendMessage(context.getChannel(), output.toString());
		}
	}

	private static final Command dumpRoles = Command.builder("internal:dumpRoles")
			.aliases("dumpRoles")
			.description("Dump all Roles ID for the current Guild.")
			.options(new HashSet<>(Arrays.asList(
					new DefaultPermissionOption(DefaultPermissionOption.Value.ANY_OWNER),
					new ScopeOption(ScopeOption.CommandScope.GUILD_CHAT))))
			.build(InternalCommands::dumpRolesImpl);

	private static void dumpRolesImpl(Context context)
	{
		if(context.getGuild().getRoles().size() == 0)
		{
			DiscordInteractionsUtil.sendMessage(context.getChannel(), "This guild has no role defined.");
			return;
		}

		StringBuilder sb = new StringBuilder("**Roles ID** for ").append(context.getGuild().getName()).append("\n")
				.append("```\n");

		int longestName = context.getGuild().getRoles().stream().max(Comparator.comparingInt(i -> i.getName().length())).get().getName().length();
		for(IRole i : context.getGuild().getRoles())
			sb.append("* ").append(StringUtils.rightPad(i.getName(), longestName)).append(" - ").append(i.getID()).append("\n");

		sb.append("```");
		DiscordInteractionsUtil.sendMessage(context.getChannel(), sb.toString());
	}
}
