package net.pingex.dcf.commands.audit;

import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.List;
import java.util.Map.Entry;

/**
 * This class contains results for an audit.
 */
public class AuditResult
{
	/**
	 * Result code for this audit
	 */
	private ResultCode opcode;

	/**
	 * Details message about what happened. Mostly used with FAIL and WARN opcodes.
	 */
	private String message;

	/**
	 * Sub audits.
	 */
	private List<Entry<IAuditComponentProvider, AuditResult>> subAuditsResults;

	/**
	 * Full constructor, with sub audits.
	 * @param opcode The result code of the operation.
	 * @param message Optional message, pass `null` for no message.
	 * @param subAuditsResults A result for each sub audit.
	 */
	public AuditResult(ResultCode opcode, String message, List<Entry<IAuditComponentProvider, AuditResult>> subAuditsResults)
	{
		this.opcode = opcode;
		this.message = message;
		this.subAuditsResults = subAuditsResults;
	}

	/**
	 * Constructor without sub audits.
	 * @param opcode The result code of the operation.
	 * @param message Optional message, pass `null` for no message.
	 */
	public AuditResult(ResultCode opcode, String message)
	{
		this.opcode = opcode;
		this.message = message;
		this.subAuditsResults = null;
	}

	/**
	 * Constructor without sub audits or a message.
	 * @param opcode The result code of the operation.
	 */
	public AuditResult(ResultCode opcode)
	{
		this.opcode = opcode;
		this.message = null;
		this.subAuditsResults = null;
	}

	/**
	 * Result code for this audit
	 */
	public ResultCode getOpcode()
	{
		return opcode;
	}

	/**
	 * Details message about what happened. Mostly used with FAIL and WARN opcodes.
	 */
	public String getMessage()
	{
		return message;
	}

	/**
	 * Sub audits, if any.
	 * @throws UnsupportedOperationException in case there is not sub audits.
	 */
	public List<Entry<IAuditComponentProvider, AuditResult>> getSubAuditsResults() throws UnsupportedOperationException
	{
		if(subAuditsResults != null)
			return subAuditsResults;
		else throw new UnsupportedOperationException();
	}

	/**
	 * Tells whether this object contains subaudits.
	 * @return `true` when this object contains subaudits, `false` otherwise
	 */
	public boolean hasSubAudits()
	{
		return subAuditsResults != null;
	}

	/**
	 * All the possible result codes for an andit.
	 */
	public enum ResultCode implements Comparable<ResultCode>
	{
		/**
		 * The audit failed.
		 */
		FAIL,

		/**
		 * The test passed with a warning (contained in the message field)
		 */
		WARN,

		/**
		 * The test passed without any issue.
		 */
		PASS,

		/**
		 * Test was ignored.
		 */
		NOOP
	}

	/**
	 * Helper class to construct a proper AuditResult.
	 */
	public static class Builder
	{
		private ResultCode opcode;
		private String message;
		private List<Entry<IAuditComponentProvider, AuditResult>> subAuditsResults = new ArrayList<>();

		public Builder setOpcode(ResultCode opcode)
		{
			this.opcode = opcode;
			return this;
		}

		public Builder setMessage(String message)
		{
			this.message = message;
			return this;
		}

		public Builder appendAuditResult(IAuditComponentProvider provider, AuditResult result)
		{
			appendAuditResult(new AbstractMap.SimpleEntry<IAuditComponentProvider, AuditResult>(provider, result));
			return this;
		}

		public Builder appendAuditResult(Entry<IAuditComponentProvider, AuditResult> entry)
		{
			subAuditsResults.add(entry);
			return this;
		}

		public List<Entry<IAuditComponentProvider, AuditResult>> getSubAuditsResults()
		{
			return subAuditsResults;
		}

		public ResultCode getOpcode()
		{
			return opcode;
		}

		public String getMessage()
		{
			return message;
		}

		public AuditResult build()
		{
			return new AuditResult(opcode, message, subAuditsResults.isEmpty() ? null : subAuditsResults);
		}
	}

	/**
	 * Fork a new builder.
	 */
	public static Builder builder()
	{
		return new Builder();
	}
}
