package net.pingex.dcf.commands.permissions;

import net.pingex.dcf.commands.Context;
import net.pingex.dcf.commands.audit.AuditResult;
import net.pingex.dcf.commands.audit.IAuditComponentProvider;
import net.pingex.dcf.commands.permissions.audit.DefaultPermissionCheck;
import net.pingex.dcf.commands.permissions.audit.UserGlobalCheck;
import net.pingex.dcf.commands.permissions.audit.UserGuildCheck;
import net.pingex.dcf.commands.permissions.audit.UserGuildRoleCheck;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Set;
import java.util.TreeSet;

/**
 * This class is used to check whether an user can run a command.
 */
public class PermissionCheck implements IAuditComponentProvider
{
	private static final ICommandPermissionsProvider CURRENT_PROVIDER = new DefaultPermissionsProvider();

	private static final Set<IAuditComponentProvider> SUB_CHECKS = new TreeSet<>(Comparator.comparingInt(IAuditComponentProvider::priority));

	static
	{
		SUB_CHECKS.addAll(Arrays.asList(
				new DefaultPermissionCheck(),
				new UserGlobalCheck(getProvider()),
				new UserGuildCheck(getProvider()),
				new UserGuildRoleCheck(getProvider())
		));
	}

	/**
	 * Get provider used to check for permissions.
	 */
	public static ICommandPermissionsProvider getProvider()
	{
		return CURRENT_PROVIDER;
	}

	@Override
	public AuditResult doAudit(Context context)
	{
		AuditResult.Builder globalResult = new AuditResult.Builder();

		for(IAuditComponentProvider i : SUB_CHECKS)
		{
			AuditResult interResult = i.doAudit(context);
			globalResult.appendAuditResult(i, interResult);

			// Check opcode has not been written yet and interResult has opcode either FAIL or PASS.
			if(globalResult.getOpcode() == null && (interResult.getOpcode() == AuditResult.ResultCode.FAIL || interResult.getOpcode() == AuditResult.ResultCode.PASS))
			{
				globalResult.setOpcode(interResult.getOpcode());
				globalResult.setMessage(interResult.getMessage());
			}
		}

		if(globalResult.getOpcode() == null)
		{
			globalResult.setOpcode(globalResult.getSubAuditsResults().get(globalResult.getSubAuditsResults().size()-1).getValue().getOpcode());
			globalResult.setMessage(globalResult.getSubAuditsResults().get(globalResult.getSubAuditsResults().size()-1).getValue().getMessage());
		}

		return globalResult.build();
	}

	@Override
	public String name()
	{
		return "Permission check";
	}

	@Override
	public String description()
	{
		return "Checks whether an user can run a command.";
	}

	@Override
	public int priority()
	{
		return 10;
	}
}
